/* -*- mode: c++ -*- */
#ifndef __UTILS_H__
#define __UTILS_H__

#include <sys/time.h>
#include <sys/types.h>
#include <inttypes.h>
#include <algorithm>
#include <gmpxx.h>

typedef uint32_t N;
typedef int32_t Z;

namespace Settings {
#include "settings.h"
}

#define TRAP asm("int $0x03\n\tnop\n\t")
typedef mpq_class QQ;
typedef mpz_class ZZ;
typedef uint64_t N64;
typedef int64_t Z64;
typedef unsigned long NNat;
typedef long ZNat;

/* returns the current time, in milliseconds */
uint64_t current_time()
{
    timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec*1000ULL + tv.tv_usec/1000ULL;
}

char* format_time(char* buf, uint64_t msec)
{
    if(msec >= 60*60*1000)
        sprintf(buf, "%uh %um",  (uint)(msec / (60*60*1000)),
                (uint)((msec % (60*60*1000)) / (60*1000)) );
    else if (msec >= 60*1000)
        sprintf(buf, "%um %us",  (uint)(msec / (60*1000)),
                (uint)((msec % (60*1000)) / 1000) );
    else
        sprintf(buf, "%us", (uint)(msec / 1000));

    return buf;
}

template<N x>
class Log2;

template<N x>
class Log2
{
public:
    static const N s_value = Log2<(x>>1)>::s_value + 1;    
};

template<>
class Log2<1>
{
public:    
    static const N s_value = 1;
};

template<>
class Log2<0>
{
public:    
    static const N s_value = 0;
};


template<N a, N b>
class Cmp
{
public:
    static const N s_max = a>b ? a : b;
    static const N s_min = a<b ? a : b;
};

template<N a, N b, N c>
class Sort3
{
public:    
    static const N s_biggest = Cmp<Cmp<a,b>::s_max, c>::s_max;
    static const N s_middle  = Cmp< Cmp<Cmp<a,b>::s_min, Cmp<a,c>::s_min>::s_max,
                                    Cmp<c,b>::s_min>::s_max;
    static const N s_smallest = Cmp<Cmp<a,b>::s_min, c>::s_min;
};

/* class formed by a 64-bit integer accumulator and a big integers, to speed up
   multiplications and divisions. The accumulator will be multiplied untill it is about
   to overflow, and each time this happens the big integer will be multiplied or diveded
   accordingly, and the accumulator will be reset. */
class QuickZ
{
public:
    ZZ zz;
    ZNat z;

    static const ZNat s_test_overflow_default =
        -((ZNat)1 << (sizeof(ZNat)*8 - Log2<Settings::d>::s_value - 1));
    static ZNat s_test_overflow;

    static void set_d(N d)
    {
        N l = 0;
        while(d>>l) l++;
        s_test_overflow = -((ZNat)1 << (sizeof(ZNat)*8 - l - 1));
    }

    QuickZ() {}

    QuickZ(const ZZ& zzz) : zz(zzz), z(1) {}

    void reset()
    {
        zz = 1;
        z = 1;
    }

    void reset(const ZZ& zzz)
    {
        zz = zzz;
        z = 1;
    }

    void reset(const ZNat& zzz)
    {
        zz = 1;
        z = zzz;
    }

    void flip_sign() { z = -z; }

    void normalize() { zz *= z; z=1; }
    
    void operator*=(const ZNat& x)
    {
        if(std::abs(z) & s_test_overflow)
        {
            zz *= z;
            z = x;
        }
        else
            z *= x;
    }

    void div_normalize() { zz /= z; z=1; }
    
    void operator/=(const ZNat& x)
    {
        if(std::abs(z) & s_test_overflow)
        {
            zz /= z;
            z = x;
        }
        else
            z *= x;
    }
};

/* computes the exponential */
ZZ exp(ZZ a, N e)
{
    ZZ retv;
    mpz_pow_ui(retv.get_mpz_t(), a.get_mpz_t(), e);
    return retv;
}

/* computes the gcd */
ZZ gcd(ZZ a, ZZ b)
{
    ZZ retv;
    mpz_gcd(retv.get_mpz_t(), a.get_mpz_t(), b.get_mpz_t());
    return retv;
}

/* computes the factorial */
ZZ factorial(N num)
{
    ZZ retv = 1;
    for(N i=2;i<=num;i++)
        retv *= i;
    return retv;
}

/* computes the sign of a permutation of len 'len'
   of numbers 0,1,...,(len-1), stored in 'perm' */
N perm_sign(N len, N* perm)
{
    N retv = 0;
    for(N i=len-1; i>0; i--)
    {
        if(perm[i] == i)
            continue;

        N j = perm[i];
        perm[i] = i;
        //printf("%u -> %u\n", i, j);
        while(j != i)
        {
            N newj = perm[j];
            perm[j] = j;
            //printf("%u -> %u\n", j, newj);
            j = newj;
            retv++;
        }
    }

    return retv & 1;
}

#endif //__UTILS_H__
